from typing import Unpack, TypedDict
from q.api import WriteQueue, ReadQueue, Queue
from pipeteer import MakeQueue, make_queues as _make_queues, PipelineQueues
from .spec import Result, workflow, Input, Preprocess, PreInput, PreResult

class PreinputPipeline:
  In = Input
  Out = PreInput
  QueueIn = ReadQueue[Input]
  QueueOut = WriteQueue[PreInput]
  Queues = PipelineQueues[Input, PreInput]

class JoinPipeline:
  In = PreResult
  Out = Result
  QueueIn = ReadQueue[PreResult]
  QueueOut = WriteQueue[Result]
  Queues = PipelineQueues[PreResult, Result]



class Workflow:
  class InternalQueues(TypedDict):
    preprocess: Preprocess.Queues
    preinput: PreinputPipeline.Queues  
    join: JoinPipeline.Queues  

  class Queues(TypedDict):
    Qin: Queue[Input]
    internal: 'Workflow.InternalQueues'

  @staticmethod
  def make_queues(make_queue: MakeQueue, output_queue: WriteQueue[Result]) -> Queues:
    return _make_queues(workflow, make_queue, output_queue) # type: ignore
  
  @staticmethod
  def artifacts(**queues: Unpack['Workflow.InternalQueues']):
    from .main import artifacts
    return artifacts(**queues)
    


from argparse import ArgumentParser

def main():
  parser = ArgumentParser()
  parser.add_argument('-b', '--base-path', required=True)
  parser.add_argument('--images', required=True)

  parser.add_argument('-p', '--port', default=8000, type=int)
  parser.add_argument('--host', default='0.0.0.0', type=str)

  args = parser.parse_args()


  import os
  from dslog import Logger
  base_path = os.path.join(os.getcwd(), args.base_path)
  images_path = os.path.join(os.getcwd(), args.images)
  
  logger = Logger.click().prefix('[GAME PREPROCESS]')
  logger(f'Running preprocessing...')
  logger(f'Images path: "{images_path}"')
  logger(f'Internal path: "{base_path}"')
  
  from typing import Sequence
  import asyncio
  from multiprocessing import Process
  from kv.fs import FilesystemKV, FilesystemAppendKV
  from kv.sqlite import SQLiteKV
  from q.kv import QueueKV
  import uvicorn
  from moveread.pipelines.game_preprocess import Workflow, Result
  from moveread.pipelines.game_preprocess.spec import Game, PreResult

  images = FilesystemKV[bytes](images_path)

  db_path = os.path.join(base_path, 'data.sqlite')

  def make_queue(path: Sequence[str], type: type):
    return QueueKV.sqlite(type, db_path, table='-'.join(['queues', *path]))
  
  Qout = make_queue(['output'], Result)
  Qs = Workflow.make_queues(make_queue, Qout)
  artifacts = Workflow.artifacts(**Qs['internal'])(
    images=images, images_path=images_path,
    games=SQLiteKV.validated(Game, db_path, 'games'),
    imgGameIds=SQLiteKV.at(db_path, 'game-ids'),
    received_imgs=FilesystemAppendKV.validated(tuple[str, PreResult], os.path.join(base_path, 'received_imgs'))
  )

  ps = {
    id: Process(target=asyncio.run, args=(f,)) for id, f in artifacts.processes.items()
  } | {
    'api': Process(target=uvicorn.run, args=(artifacts.api,), kwargs={'host': args.host, 'port': args.port})
  }
  for id, p in ps.items():
    p.start()
    logger(f'Process "{id}" started at PID {p.pid}')
  for p in ps.values():
    p.join()
  

if __name__ == '__main__':
  import sys
  import os
  os.chdir('/home/m4rs/mr-github/rnd/data/moveread-pipelines/backend/3.game-preprocess/')
  sys.argv.extend('-b demo --images demo/images/'.split(' '))
  main()


from typing import Callable
from uuid import uuid4
from kv.api import KV
from q.api import WriteQueue
from dslog import Logger
from scoresheet_models import ModelID
from moveread.core import CoreAPI, Game
from moveread.pipelines.game_preprocess import Input

async def input_core(
  core: CoreAPI, Qin: WriteQueue[Input],
  *, images: KV[bytes],
  model_fn: Callable[[Game], ModelID] = lambda *_: 'llobregat23',
  num_games: int | None = None, shuffle: bool = True,
  logger = Logger.rich().prefix('[CORE INPUT]')
):
  """Input all images from `core` into `Qin` tasks
  - Actually, only images with `version == 0`
  - `model_fn`: determines the scoresheet model of each task
  - `state_fn`: determines an arbitrary tuple of JSON-serializable data to attach to each task
  """
  games = list((await core.games.keys()).unsafe())
  if shuffle:
    import random
    random.shuffle(games)
  for gameId in games[:num_games]:
    game = (await core.games.read(gameId)).unsafe()
    imgs = []
    for imgId, image in game.images:
      if imgId.version == 0:
        id = str(imgId)
        url = f'{id}/original_{uuid4()}.jpg'
        img = (await core.blobs.read(image.url)).unsafe()
        (await images.insert(url, img)).unsafe()
        imgs.append(url)

    task = Input(model=model_fn(game), imgs=imgs)
    await Qin.push(gameId, task)
    logger(f'Inputted task "{gameId}". Task:', task)


from .spec import Input, Result, workflow, Game
from .integrations import input_core
from .main import Params, Artifacts
from .spec_codegen import Workflow

__all__ = [
  'Input', 'Result', 'workflow', 'Game',
  'Params', 'Artifacts', 'Workflow',
  'input_core'
]